/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2012
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2012
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CStratThreadedUp.cpp
 *        Application phase aware communication strategy.
 *
 * Provides the interface functionality with an implementation
 * that tries to avoid any communication. during phases in which the
 * application is communicating.
 *
 * @author Tobias Hilbrich
 * @date 24.08.2009
 */

#include "CStratThreadedUp.h"
#include "GtiMacros.h"
#include "GtiApi.h"

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <pnmpimod.h>
#include <stdlib.h>
#include <string.h>

using namespace gti;

mGET_INSTANCE_FUNCTION(CStratThreadedUp)
mFREE_INSTANCE_FUNCTION(CStratThreadedUp)
mPNMPI_REGISTRATIONPOINT_FUNCTION(CStratThreadedUp)
mCOMM_STRATEGY_UP_RAISE_PANIC(CStratThreadedUp)

//=============================
// my_buf_free_function
//=============================
GTI_RETURN my_buf_free_function (void* free_data, unsigned long num_bytes, void* buf)
{
	delete [] (unsigned long*) buf;
	return GTI_SUCCESS;
}

//=============================
// CStratThreadedUp
//=============================
CStratThreadedUp::CStratThreadedUp (const char* instanceName)
    : ModuleBase<CStratThreadedUp, CStratUpQueue> (instanceName),
      CStratThreadedAggregator (&protocol),
      myAggregationAllowed (true)
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //A Comm Strategy either uses a regular communication protocol or a regular and a
    //fall-back one.
    assert ((subModInstances.size() == 1) || (subModInstances.size() == 2));

    protocol = (I_CommProtocol*) subModInstances[0];

    //Set up own data
    myRequest = 0xFFFFFFFF;
    myGotPing = false;

    //Set up buffers immediately for our single communication channel
    myCurAggregateBufs.resize(1);
    myCommBufs.resize(1);
    myCurrAggregateLens.resize(1);
    myCurAggregateBufs[0] = NULL;
    myCommBufs[0] = NULL;
    myCurrAggregateLens[0] = 0;
    swap(0);
}

//=============================
// ~CStratThreadedUp
//=============================
CStratThreadedUp::~CStratThreadedUp (void)
{
    /*Destroy sub modules*/
    if (protocol)
        destroySubModuleInstance ((I_Module*)protocol);
    protocol = NULL;

#ifdef GTI_DEBUG
    //DEBUG OUT
    std::cout << "MaxReqs: " << myMaxNumReqs << std::endl;
#endif
}

//=============================
// shutdown
//=============================
GTI_RETURN CStratThreadedUp::shutdown (
        GTI_FLUSH_TYPE flush_behavior,
        GTI_SYNC_TYPE sync_behavior)
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==shutdown
    if (flush_behavior == GTI_FLUSH)
    {
        flush ();

        //remove outstanding requests
        while (!myRequests.empty())
        {
            AggRequestInfo req = myRequests.front();
            protocol->wait_msg(req.request,NULL,NULL);
            delete []req.buf;
            myRequests.pop_front();
        }
    }

    // Remove protocol requests, to avoid sync msgs beeing catched by
    // outstanding requests !
    protocol->removeOutstandingRequests ();

    if (sync_behavior == GTI_SYNC)
    {
        unsigned long buf[2] = {0,0};
        unsigned long length;

        //Receive the ping (if not already received)
        while ((buf[0] != myTokenShutdownSync) && (!myGotPing))
        {
            protocol->recv(buf,sizeof(unsigned long)*2,&length,0,NULL);
            assert(length == sizeof(unsigned long)*2);

            if (buf[0] != myTokenShutdownSync)
            {
                //TODO give a warning here, received some other outstanding
                //     message even though this strategy is being closed
                //Receive the message, and discard it
                std::cerr << "WARNING: In shutdown sync (StrategyUp) "
                        << "received an outstanding message!"
                        << std::endl;
                unsigned long *temp_buf;
                length = buf[1];
                if (length % sizeof(unsigned long)) length = ((length/sizeof(unsigned long))+1)*sizeof(unsigned long);
                temp_buf = new unsigned long[length/sizeof(unsigned long)];
                protocol->recv(temp_buf,buf[1],&length,0, NULL);
                delete [] temp_buf;
            }
        }

        //Send the pong
        buf[0] = myTokenShutdownSync;
        buf[1] = 0; //of no importance
        protocol->ssend(buf,sizeof(unsigned long)*2,0);
    }

    //Shutdown the protocols
    protocol->shutdown();
    return GTI_SUCCESS;
}

//=============================
// send
//=============================
GTI_RETURN CStratThreadedUp::send (
        void* buf,
        unsigned long num_bytes,
        void* free_data,
        GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        AddToQueue (buf, num_bytes, free_data, buf_free_function);
        return GTI_SUCCESS;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==send

    //DEBUG
    //std::cout << "SEND: [" << num_bytes << "]" << "id=" << ((unsigned long*) buf)[0] << std::endl;
    //ENDDEBUG

    //Is it a long message ?
    if (num_bytes + 3*sizeof (unsigned long) > BUF_LENGTH) //3 -> message token + num records + one record length
    {
        return send_long_message (0, buf, num_bytes, free_data, buf_free_function);
    }

    //Does it still fit into the aggregate buffer ?
    if (myCurrAggregateLens[0] + num_bytes + sizeof(unsigned long) > BUF_LENGTH)
    {
        //Its too long ...
        unsigned long comm_buf_length = myCurrAggregateLens[0];
        swap(0); //finish up outstanding communication of last buffer and swap buffers

        //DEBUG
        //std::cout << "SEND AGGREGATE: [" << comm_buf_length << "]" << std::endl;
        //ENDDEBUG

        sendCommBuf (false, comm_buf_length, 0);
    }

    //Add the new message to the aggregate buffer
    unsigned long startIndex = (myCurrAggregateLens[0]/sizeof(unsigned long));
    ((unsigned long*)myCurAggregateBufs[0])[1] = ((unsigned long*)myCurAggregateBufs[0])[1] + 1; //increase count of msgs
    ((unsigned long*)myCurAggregateBufs[0])[startIndex] = num_bytes;
    myCurrAggregateLens[0] += sizeof (unsigned long);
    memmove (myCurAggregateBufs[0]+myCurrAggregateLens[0], buf, num_bytes);
    myCurrAggregateLens[0] += num_bytes;

    //pad to a multiple of "sizeof(unsigned long)"
    if (myCurrAggregateLens[0] % sizeof(unsigned long))
        myCurrAggregateLens[0] += sizeof(unsigned long) - myCurrAggregateLens[0] % sizeof(unsigned long);

    //remove padding if buffer becomes too long due to that
    if (myCurrAggregateLens[0] > BUF_LENGTH)
        myCurrAggregateLens[0] = BUF_LENGTH;

    //free given buffer
    buf_free_function (free_data, num_bytes, buf);

    //Are we still allowed to aggregate?
    if (!myAggregationAllowed)
        flush ();

    return GTI_SUCCESS;
}

//=============================
// test
//=============================
GTI_RETURN CStratThreadedUp::test (
        int* out_flag,
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==test
    if (myRequest == 0xFFFFFFFF)
    {
        protocol->irecv(myBuf,sizeof(unsigned long)*2,&myRequest,0);
    }

    int completed;
    unsigned long numbytes, channel;

    protocol->test_msg(myRequest,&completed,&numbytes,&channel);

    if (completed)
    {
        //Save request as used up
        myRequest = 0xFFFFFFFF;

        assert (numbytes == sizeof(unsigned long)*2); //has to be a token,length message

        //Did we receive the ping used for ping-pong shutdown sync ?
        if (myBuf[0] == myTokenShutdownSync)
        {
            assert (myGotPing == false);
            myGotPing = true;
            //Retry, maybe another message available ...
            return test (out_flag, out_num_bytes, out_buf, out_free_data, out_buf_free_function);
        }

        assert (myBuf[0] == myTokenMessage); //has to be a message

        unsigned long* buf;
        unsigned long  buf_size, recv_size;

        if (myBuf[1] % sizeof(unsigned long) != 0)
            buf_size = (myBuf[1]/sizeof(unsigned long))+1;
        else
            buf_size = (myBuf[1]/sizeof(unsigned long));

        buf = new unsigned long[buf_size];

        protocol->recv(buf,myBuf[1],&recv_size,channel,NULL);

        assert (recv_size == myBuf[1]); //must be the size specified in first message

        //save values
        *out_flag = 1;
        *out_num_bytes = recv_size;
        *out_buf = buf;
        *out_free_data = NULL;
        *out_buf_free_function = my_buf_free_function;
    }
    else
    {
        //No message received
        *out_flag = 0;
    }

    return GTI_SUCCESS;
}

//=============================
// wait
//=============================
GTI_RETURN CStratThreadedUp::wait (
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==wait
    unsigned long numbytes, channel;

    if (myRequest != 0xFFFFFFFF)
    {
        protocol->wait_msg(myRequest,&numbytes,&channel);
    }
    else
    {
        protocol->recv(myBuf,sizeof(unsigned long)*2,&numbytes,0,&channel);
    }

    //Save request as used up
    myRequest = 0xFFFFFFFF;

    //Sanity
    assert (numbytes == sizeof(unsigned long)*2); //has to be a token,length message

    //Did we receive the ping used for ping-pong shutdown sync ?
    if (myBuf[0] == myTokenShutdownSync)
    {
        assert (myGotPing == false);
        myGotPing = true;
        //Retry, maybe another message available ...
        return wait (out_num_bytes, out_buf, out_free_data, out_buf_free_function);
    }

    //Regular message
    assert (myBuf[0] == myTokenMessage); //has to be a message

    unsigned long* buf;
    unsigned long  buf_size, recv_size;

    if (myBuf[1] % sizeof(unsigned long) != 0)
        buf_size = (myBuf[1]/sizeof(unsigned long))+1;
    else
        buf_size = (myBuf[1]/sizeof(unsigned long));

    buf = new unsigned long[buf_size];

    protocol->recv(buf,myBuf[1],&recv_size,channel, NULL);

    assert (recv_size == myBuf[1]); //must be the size specified in first message

    //save values
    *out_num_bytes = recv_size;
    *out_buf = buf;
    *out_free_data = NULL;
    *out_buf_free_function = my_buf_free_function;

    return GTI_SUCCESS;
}

//=============================
// raisePanic
//=============================
GTI_RETURN CStratThreadedUp::raisePanic (void)
{
    //1) Flush
    flush ();

    //2) Disable future aggregation
    myAggregationAllowed = false;

    //3) Create the GTI internal panic event
    gtiRaisePanicP f;
    if (getWrapperFunction("gtiRaisePanic", (GTI_Fct_t*) &f) == GTI_SUCCESS)
    {
        //Raise the panic
        (*f) ();
    }
    else
    {
        std::cerr << "MUST internal error: could not find the creation function for the GTI internal event \"gtiRaisePanic\", this should never happen (" << __FILE__ << ":" << __LINE__ << ")." << std::endl;
        return GTI_ERROR;
    }

    return GTI_SUCCESS;
}

//=============================
// flush
//=============================
GTI_RETURN CStratThreadedUp::flush (
        void
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==flush
    int completed;

    //Test for completion of existing send operations
    unsigned long old_len = myCurrAggregateLens[0];
    if (old_len > 2*sizeof(unsigned long))
    {
        swap(0); //finish sending old buffer and swap buffers

        //DEBUG
        //std::cout << "SEND SHUTDOWN AGGREGATE: [" << old_len << "]" << std::endl;
        //ENDDEBUG

        sendCommBuf (false, old_len, 0);
    }

    return GTI_SUCCESS;
}

//=============================
// flushAndSetImmediate
//=============================
GTI_RETURN CStratThreadedUp::flushAndSetImmediate (void)
{
    //1) Flush
    flush ();

    //2) Set immediate
    myAggregationAllowed = false;

    return GTI_SUCCESS;
}

/*EOF*/
